#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>
#include <signal.h>

#include "probedev.h"

#define HOTPLUG_BIN		"/sbin/hotplug"
#define HOTPLUG_PROCPATH	"/proc/sys/kernel/hotplug"
#define HOTPLUG_PIPE		"/tmp/hotplug_comm_pipe"

#define BUFFER_SIZE		512
char g_hprd_buff[BUFFER_SIZE+1];
size_t g_hprd_n;
int g_hprd_fd;


void exec_hotplug( int argc, char **argv ) {
	char *action;
	char *devpath;
	int i, fd;
	static int maxtries = 5;
	
	action = getenv("ACTION");
	if (!action || strcmp(action, "add")) return;

	devpath = getenv("DEVPATH");
	if (!devpath) return;

	/* Try several times to write the device */
	i = 0;
	do {
		if ((fd=open(HOTPLUG_PIPE, O_WRONLY)) == -1)
			continue;
		if (flock(fd, LOCK_EX) != 0)
			continue;
		write(fd, devpath, strlen(devpath));
		write(fd, "\n", 1);
		flock(fd, LOCK_UN);
		close(fd);
		break;
	} while ( i++ < maxtries && (sleep(1) || 1));	/* If we're continuing, sleep */
}

void init_hotplug( void ) {
	int fd;

	g_hprd_n = 0;

	unlink(HOTPLUG_PIPE);
	if (mknod(HOTPLUG_PIPE, S_IFIFO, 0) != 0) {
		fprintf(stderr, "Unable to create hotplug pipe: %s\n", strerror(errno));
		exit(1);
	}
	if (chmod(HOTPLUG_PIPE, 0666) != 0) {
		fprintf(stderr, "Unable to chmod hotplug pipe: %s\n", strerror(errno));
		exit(1);
	}

	if ((g_hprd_fd=open(HOTPLUG_PIPE, O_RDONLY | O_NONBLOCK)) == -1) {
		fprintf(stderr, "Unable to open hotplug pipe: %s\n", strerror(errno));
		exit(1);
	}

	if ((fd=open(HOTPLUG_PROCPATH, O_WRONLY | O_TRUNC)) == -1) {
		fprintf(stderr, "Unable to open %s: %s\n", HOTPLUG_PROCPATH, strerror(errno));
		exit(1);
	}
	if (write(fd, HOTPLUG_BIN, strlen(HOTPLUG_BIN)) != strlen(HOTPLUG_BIN)) {
		fprintf(stderr, "Unable to set hotplug executable: %s\n", strerror(errno));
		exit(1);
	}
	close(fd);

}

void cleanup_hotplug( void ) {
	int fd;

	close(g_hprd_fd);

	unlink(HOTPLUG_PIPE);	
	if ((fd=open(HOTPLUG_PROCPATH, O_WRONLY | O_TRUNC)) == -1) {
		fprintf(stderr, "Unable to open %s: %s\n", HOTPLUG_PROCPATH, strerror(errno));
		exit(1);
	}
	if (write(fd, "", 1) != 1) {
		fprintf(stderr, "Unable to set hotplug executable: %s\n", strerror(errno));
		exit(1);
	}
	close(fd);

	/* We ignore SIGTERM, but the child hotplug processes don't */
	kill(-1, SIGTERM);
}

int check_buffer( void ) {
	int i;
	char *s;

	for (i=0; i<=g_hprd_n && g_hprd_buff[i] != '\n'; i++);
	if (i > g_hprd_n) return 0;
	g_hprd_buff[i] = '\0';

	s = (char *)malloc(strlen(g_hprd_buff) + 5);
	sprintf(s, "/sys%s", g_hprd_buff);
	probe_sys_path(s);
	free(s);
	
	i++;
	if (g_hprd_n-i > 0) {
		memmove(g_hprd_buff, (char *)g_hprd_buff + i, g_hprd_n-i);
		g_hprd_n -= i;
	} else {
		g_hprd_n = 0;
	}

	return 1;
}

int hotplug_process( void ) {
	char bfr[64];
	ssize_t n;
	int i;
	int rc = 0;

	if (flock(g_hprd_fd, LOCK_EX) != 0) {
		printf("ERROR Acquiring lock on hotplug pipe: %s\n", strerror(errno));
		return 0;
	}

	while ((n=read(g_hprd_fd, bfr, sizeof(bfr)-1)) > 0) {
		bfr[n>=sizeof(bfr) ? sizeof(bfr)-1 : n] = '\0';
		/* Can fit it all, just copy */
		if (g_hprd_n+n <= BUFFER_SIZE) {
			memcpy((char *)g_hprd_buff + g_hprd_n, bfr, n);
			g_hprd_n += n;
		} else {
			/* Fill entirely */
			if (g_hprd_n < BUFFER_SIZE) {
				i = BUFFER_SIZE-g_hprd_n;
				memcpy((char *)g_hprd_buff + g_hprd_n, bfr, i);
			} else {
				i = 0;
			}


			/* Shift out what we need to */
			memmove(g_hprd_buff, (char *)g_hprd_buff + (n-i), BUFFER_SIZE-(n-i));
			/* Put remaining */
			memcpy((char *)g_hprd_buff + g_hprd_n - (n-i), (char *)bfr + i, n-i);
		}
		/* Keep checking until we run out of things */
		while ((i=check_buffer()) != 0) rc |= i;
	}
	flock(g_hprd_fd, LOCK_UN);
	return rc;
}
