#include <linux/version.h>
#include <linux/loop.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/mount.h>
#include <errno.h>
#include <dirent.h>
#include <utime.h>

#include "setuproot.h"
#include "filesys.h"
#include "archive_extract.h"
#include "decompress_bunzip2.h"

/* Loop device to use for the root filesystem */
#define LOOP_MAJOR          7
#define LOOP_ROOT_MINOR     0

/* NOTE: These all need to be in /mnt, since they all use the initrd's
   symlink to the tmpfs /newroot/mnt */

#define ROOT_DEV        "/mnt/rootdev"      /* Root device node */
#define ROOT_LOOP       "/mnt/rootloop"     /* Root loopback device node */
#define ROOT_DEV_MNT    "/mnt/rootdevmnt"   /* Root device mount point */
#define ROOT_IMG_MNT    "/mnt/rootimg"      /* Root image mount point */
#define OVR_DEV         "/mnt/overridedev"  /* Override device node (if any */
#define OVR_MNT         "/mnt/override"     /* Override device mount point 
                                               (possibly a symlink to 
                                               ROOT_DEV_MNT if they're the 
                                               same device */

#define OVERRIDE_LIST   "/var/override"     /* List of file that are 
                                                overriden */


/* Directories to create on the new root as empty */
static char *g_empty_dirs[] = {
    "/mnt",
    "/proc",
    "/sys",
    "/tmp",
    ROOT_DEV_MNT,
    ROOT_IMG_MNT,
};

/* To have their toplevel contents copied */
static char *g_copy_dirs[] = {
    /* System config */
    "/etc",
    "/etc/X11",
    "/etc/network",
    "/etc/default",
    "/etc/udev",
    "/etc/udev/rules.d",

    /* Generally unchanging */
    "/bin",
    "/sbin",
    "/dev",
    "/root",
    "/usr",
    "/lib",

    /* Home directories */
    "/home",
    "/home/aerosol",
    "/home/cpd",
    "/home/cpd/.dosemu",
    "/home/cpd/.dosemu/drives",
    "/home/cpd/.juniper_networks",
    "/home/cpd/.juniper_networks/network_connect",

    /* idesk stuff */
    "/etc/idesk",
    "/etc/idesk/desktop",

    /* /var stuff */
    "/var",
    "/var/lock",
    "/var/run",
    "/var/run/ntp",
    "/var/lib",
    "/var/cache",
    "/var/cache/hald",
    "/var/lib/dhcp",
    "/var/lib/dbus",
    "/var/lib/urandom",
    "/var/lib/xkb",
    "/var/lib/ntp",
    "/var/lib/samba",
    "/var/log",
    "/var/spool",
    "/var/spool/cron",
    "/var/spool/cron/crontabs",
    "/var/dosneph",
    "/var/tmp",

    /* /aer tree */
    "/aer",
    "/aer/db",
    "/aer/db/.cache",
    "/aer/db/stn",
    "/aer/prg",
    "/aer/prg/bash",
    "/aer/prg/cpd",
    "/aer/prg/cpd/etc",
    "/aer/prg/cpd2",
    "/aer/prg/cpd2/etc",
};


#define BUFFER_SIZE     8192
char *g_copy_buffer = NULL;

static void clone_flags( char *file, struct stat *sb ) {
     struct utimbuf ut;
     
    if (lchown(file, sb->st_uid, sb->st_gid) != 0) {
        fprintf(stderr, "Unable to set ownership of %s: %s\n", file, strerror(errno));
        exit(1);
    }
    
    /* Can't do anything else to links */
    if (S_ISLNK(sb->st_mode)) return;

    if (chmod(file, sb->st_mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISUID|S_ISGID|S_ISVTX)) != 0) {
        fprintf(stderr, "Unable to set premissions of %s: %s\n", file, strerror(errno));
        exit(1);
    }

    if (S_ISREG(sb->st_mode)) {
        ut.actime = sb->st_atime;
        ut.modtime = sb->st_mtime;
        if (utime(file, &ut) != 0) {
            fprintf(stderr, "Unable to set access/modify time of %s: %s\n", file, strerror(errno));
            exit(1);
        }
    }
}

static void copy_contents( char *src, char *dest ) {
    int fdin, fdout;
    ssize_t nrd;

    if ((fdin=open(src, O_RDONLY)) < 0) {
        fprintf(stderr, "Unable to open %s for reading: %s\n", src, strerror(errno));
        exit(1);
    }
    if ((fdout=open(dest, O_WRONLY|O_CREAT|O_TRUNC)) < 0) {
        fprintf(stderr, "Unable to open %s for writing: %s\n", dest, strerror(errno));
        exit(1);
    }

    while ((nrd=read(fdin, g_copy_buffer, BUFFER_SIZE)) > 0) {
        if (write(fdout, g_copy_buffer, nrd) != nrd) {
            fprintf(stderr, "Error writing to file %s: %s\n", dest, strerror(errno));
            exit(1);
        }
    }
    if (nrd < 0) {
        fprintf(stderr, "Error reading from file %s: %s\n", src, strerror(errno));
        exit(1);
    }

    close(fdin);
    close(fdout);
}

static void root_mkdir( char *d ) {
    char *dirloc = (char *)malloc(strlen(d) + strlen(ROOT_LOC) + 1);
    sprintf(dirloc, ROOT_LOC "%s", d);
    if (mkdir(dirloc, 0755) != 0) { 
        fprintf(stderr, "Unable to create directory %s: %s\n", dirloc, strerror(errno)); 
        exit(1); 
    }
    free(dirloc);
}

static void clone_file( char *filesrc, char *filedest, struct stat *sb ) {
    if (S_ISREG(sb->st_mode)) {
        copy_contents( filesrc, filedest );
    } else if (S_ISLNK(sb->st_mode)) {
        char lnkbuff[256];
        int n;
        if ((n=readlink(filesrc, lnkbuff, sizeof(lnkbuff)-1)) == -1) {
            fprintf(stderr, "Unable to read link %s: %s\n", filesrc, strerror(errno));
            exit(1);
        }
        lnkbuff[n] = '\0';
        if (symlink(lnkbuff, filedest) != 0) {
            fprintf(stderr, "Unable to create symlink at %s (%s): %s\n", filedest, lnkbuff, strerror(errno));
            exit(1);
        }
    } else if (S_ISCHR(sb->st_mode) || S_ISBLK(sb->st_mode) || S_ISFIFO(sb->st_mode) || S_ISSOCK(sb->st_mode)) {
        mode_t mt;
        dev_t dt;
        if (S_ISCHR(sb->st_mode))
            mt = S_IFCHR;
        else if (S_ISBLK(sb->st_mode))
            mt = S_IFBLK;
        else if (S_ISSOCK(sb->st_mode))
            mt = S_IFIFO;
        else
            mt = S_IFIFO;
        if (S_ISCHR(sb->st_mode) || S_ISBLK(sb->st_mode))
            dt = sb->st_rdev;
        else
            dt = 0;
        if (mknod(filedest, mt, dt) != 0) {
            fprintf(stderr, "Unable to create special file %s: %s\n", filedest, strerror(errno));
            exit(1);
        }
    }
    clone_flags( filedest, sb );
}

static void copy_dir_contents( char *srcloc, char *dirloc, char *basedir ) {
    struct stat sb;
    char *filesrc = NULL;
    char *filedest = NULL;
    DIR *d;
    struct dirent *de;

    if ((d=opendir(srcloc)) == NULL) {
        fprintf(stderr, "Unable to open directory %s: %s\n", srcloc, strerror(errno));
        exit(1);
    }

    while ((de=readdir(d)) != NULL) {
        if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, "..")) continue;
        if (filesrc) free(filesrc);
        filesrc = (char *)malloc(strlen(srcloc) + strlen(de->d_name) + 2);
        sprintf(filesrc, "%s/%s", srcloc, de->d_name);
        if (lstat(filesrc, &sb) != 0) {
            fprintf(stderr, "Unable to stat %s: %s\n", filesrc, strerror(errno));
            exit(1);
        }

        if (filedest) free(filedest);
        filedest = (char *)malloc(strlen(dirloc) + strlen(de->d_name) + 2);
        sprintf(filedest, "%s/%s", dirloc, de->d_name); 

        /* Don't recurse, instead just point to the image directory */
        if (S_ISDIR(sb.st_mode)) {
            /* filesrc will always have more room, so just reuse */
            sprintf(filesrc, ROOT_IMG_MNT "%s/%s", basedir, de->d_name);
            if (symlink(filesrc, filedest) != 0) {
                fprintf(stderr, "Unable to create symlink %s (%s): %s\n", filedest, filesrc, strerror(errno));
                exit(1);
            }
            if (lchown(filedest, sb.st_uid, sb.st_gid) != 0) {
                fprintf(stderr, "Unable to set ownership of %s: %s\n", filedest, strerror(errno));
                exit(1);
            }
            continue;
        }

        clone_file( filesrc, filedest, &sb );
    }

    if (filesrc) free(filesrc);
    if (filedest) free(filedest);
    closedir(d);
}

static void copy_dir( char *d ) {
    char *dirloc = (char *)malloc(strlen(d) + strlen(ROOT_LOC) + 1);
    char *srcloc = (char *)malloc(strlen(d) + strlen(ROOT_LOC) + strlen(ROOT_IMG_MNT) + 1); 
    struct stat sb; 
    
    sprintf(dirloc, ROOT_LOC "%s", d);
    sprintf(srcloc, ROOT_LOC ROOT_IMG_MNT "%s", d);

    rmdir(dirloc);  
    unlink(dirloc);
    if (stat(srcloc, &sb) != 0) {
        fprintf(stderr, "Unable to stat %s: %s\n", srcloc, strerror(errno));
        exit(1);
    }
    if (!S_ISDIR(sb.st_mode)) {
        fprintf(stderr, "File %s is not a directory\n", srcloc);
        exit(1);
    }
    if (mkdir(dirloc, 0700) != 0) {
        fprintf(stderr, "Unable to create %s: %s\n", dirloc, strerror(errno));
        exit(1);
    }
    clone_flags(dirloc, &sb);
    
    copy_dir_contents(srcloc, dirloc, d);

    free(dirloc);   
    free(srcloc);
}

#define NDOTS       25
#define DPERPCT     5
void setup_root_filesystem( int major, int minor, char *img ) {
    struct loop_info li;
    int i, dfd, ffd, didx;
    double dlt;
    dev_t dt;

    g_copy_buffer = (char *)malloc(BUFFER_SIZE);

    printf("\nSetting up baseline filesystem from %d:%d%s\n", major,minor,img);

    /* Setup base image */
    for (i=0; i<sizeof(g_empty_dirs)/sizeof(g_empty_dirs[0]); i++) {
        root_mkdir(g_empty_dirs[i]);
    }

    /* Set /tmp */
    if (chmod(ROOT_LOC "/tmp", 01777) != 0) {
        fprintf(stderr, "Unable to set premissions of /tmp: %s\n", strerror(errno));
        exit(1);
    }

    printf("  Acquiring filesystem image...\n");

    /* Create device node and mount it */
    dt = ((major & 0xFF) << 8) | (minor & 0xFF);    /* FIXME */
    if (mknod(ROOT_LOC ROOT_DEV, S_IFBLK, dt) != 0 || 
            chmod(ROOT_LOC ROOT_DEV, 0600) != 0) {
        fprintf(stderr, "Unable to create root device node: %s\n", strerror(errno));
        exit(1);
    }
    /* NOTE: Mount is done with a path that's on the initrd but is
       symlinked to the right place.  This is done so once we
       pivot_root, /proc/mtab is still correct */
    if (!auto_fs_mount(ROOT_DEV, ROOT_DEV_MNT, MS_NOATIME|MS_NODIRATIME)) {
        if (!auto_fs_mount(ROOT_DEV, ROOT_DEV_MNT, MS_NOATIME|MS_NODIRATIME|MS_RDONLY)) {
            fprintf(stderr, "Unable to mount root device: %s\n", strerror(errno));
            exit(1);
        }
    }

    /* Set this here, so we don't need extra */
    memset(&li, 0, sizeof(li));
    snprintf(li.lo_name, LO_NAME_SIZE, ROOT_LOC ROOT_DEV_MNT "/%s", img);
    li.lo_offset = 0;

    /* Open image file */
    if ((ffd=open(li.lo_name, O_RDONLY)) < 0) {
        fprintf(stderr, "Unable to open root image: %s\n", strerror(errno));
        exit(1);
    }

    /* Create loop device */
    dt = (LOOP_MAJOR << 8) | LOOP_ROOT_MINOR;    /* FIXME */
    if (mknod(ROOT_LOC ROOT_LOOP, S_IFBLK, dt) != 0 || 
            chmod(ROOT_LOC ROOT_LOOP, 0600) != 0) {
        fprintf(stderr, "Unable to create loopback device for root filesystem: %s\n", 
                strerror(errno));
        exit(1);
    }
    if ((dfd=open(ROOT_LOC ROOT_LOOP, O_RDONLY)) < 0) {
        fprintf(stderr, "Unable to open loopback device for root filesystem: %s\n", 
                strerror(errno));
    }

    /* Tie loop device to image file */
    if (ioctl(dfd, LOOP_SET_FD, ffd) != 0) {
        fprintf(stderr, "Unable to set loop FD: %s\n", strerror(errno));
        exit(1);
    }
    if (ioctl(dfd, LOOP_SET_STATUS, &li) != 0) {
        fprintf(stderr, "Unable to set loop status: %s\n", strerror(errno));
        exit(1);
    }
    close(dfd);
    close(ffd);

    /* Mount loop device */
    if (mount(ROOT_LOOP, ROOT_IMG_MNT, "squashfs", MS_RDONLY, NULL) != 0) {
        fprintf(stderr, "Unable to mount root loop file: %s\n", strerror(errno));
        exit(1);
    }
    
    
    printf("  Copying image:  0%%"); fflush(stdout);
    for (didx=1,i=0; i<sizeof(g_copy_dirs)/sizeof(g_copy_dirs[0]); i++) {
        dlt = ((double)i/(double)(sizeof(g_copy_dirs)/sizeof(g_copy_dirs[0]))) * (double)NDOTS;
        while (didx <= (int)(dlt)) {
            if (didx && didx < NDOTS && (didx % DPERPCT) == 0) {
                printf("%d%%", (int)(((double)didx/(double)NDOTS)*100.0));
            } else {
                printf(".");
            }
            fflush(stdout);
            didx++;
        }

        copy_dir(g_copy_dirs[i]);
    }
    printf("100%%\n");
    
    
    printf("Done creating baseline filesystem\n\n");
}

static void move_to_ram( char *dir, int domkdir ) {
    struct stat sb;
    char lnkbuff[512];
    int lnklen;
    DIR *d;
    struct dirent *de;
    char *nf = NULL;

    /* Stat it */
    if (lstat(dir, &sb) != 0) {
        if (errno == ENOENT) {
            if (domkdir) {
                mkdir(dir, 0775);
            }
            return;
        }
        fprintf(stderr, "Error accessing directory %s: %s\n", dir, strerror(errno));
        exit(1);
    }
    
    /* Already a real file, thus already in ram */
    if (!S_ISLNK(sb.st_mode))
        return;

    /* Get link contents */
    if ((lnklen=readlink(dir, lnkbuff, sizeof(lnkbuff)-1)) == -1) {
        fprintf(stderr, "Unable to read link %s: %s\n", dir, strerror(errno));
        exit(1);
    }
    lnkbuff[lnklen-1] = '\0';

    /* This is a link, but doesn't point to the root image, do nothing */
    if (strncmp(lnkbuff, ROOT_IMG_MNT, strlen(ROOT_IMG_MNT))) {
        return;
    }

    /* Now stat the real file */
    if (stat(dir, &sb) != 0) {
        fprintf(stderr, "Error accessing file %s: %s\n", dir, strerror(errno));
        exit(1);
    }

    /* This isn't a directory, thus we can just remove it */
    if (!S_ISDIR(sb.st_mode)) {
        if (unlink(dir) != 0) {
            fprintf(stderr, "Unable to remove file %s: %s\n", dir, strerror(errno));
            exit(1);
        }
        return;
    }

    /* This IS a directory, we have to copy the create all the links */
    if ((d=opendir(dir)) == NULL) { /* This will open the "real" directory, not the symlinked one */
        fprintf(stderr, "Unable to open directory %s: %s\n", dir, strerror(errno));
        exit(1);
    }

    /* Remove the symlink */
    if (unlink(dir) != 0) {
        fprintf(stderr, "Unable to remove symlink %s: %s\n", dir, strerror(errno));
        exit(1);
    }

    /* Create real directory */
    if (mkdir(dir, sb.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISUID|S_ISGID|S_ISVTX)) != 0) {
        fprintf(stderr, "Unable to create directory %s: %s\n", dir, strerror(errno));
        exit(1);
    }
    if (chmod(dir, sb.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISUID|S_ISGID|S_ISVTX)) != 0 || chown(dir, sb.st_uid, sb.st_gid) != 0) {
        fprintf(stderr, "Error setting directory premissions: %s: %s\n", dir, strerror(errno));
        exit(1);
    }

    /* Link directory contents */
    while ((de=readdir(d)) != NULL) {
        if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, "..")) continue;

        nf = (char *)realloc(nf, strlen(dir) + strlen(de->d_name) + strlen(ROOT_LOC) + strlen(ROOT_IMG_MNT) + 2);
        /* Stat the orginal file and set the link to be owned by the
           right user, strip root path from dir */
        sprintf(nf, ROOT_LOC ROOT_IMG_MNT "%s/%s", dir + strlen(ROOT_LOC), de->d_name);
        if (lstat(nf, &sb) != 0) {
            fprintf(stderr, "Unable to stat original file %s: %s\n", nf, strerror(errno));
            exit(1);
        }
        
        /* Put the location in this buffer (remove the leading path
           from nf, so it points to the right spot */
        sprintf(nf, "%s/%s", dir, de->d_name);
        snprintf(lnkbuff, sizeof(lnkbuff), ROOT_IMG_MNT "%s", nf + strlen(ROOT_LOC));
        if (symlink(lnkbuff, nf) != 0) {
            fprintf(stderr, "Unable to create symlink %s (%s): %s\n", nf, lnkbuff, strerror(errno));
            exit(1);
        }
        if (lchown(nf, sb.st_uid, sb.st_gid) != 0) {
            fprintf(stderr, "Unable to set link ownership %s (%s): %s\n", nf, lnkbuff, strerror(errno));
            exit(1);
        }
    }

    if (nf) free(nf);
    closedir(d);
}

int g_override_list_target;

static int extract_filter( int fdin, OVRDA_AllHeader *hdr, char *filename ) {
    char *basename = filename;
    char *p;
    int n;
    
    /* Take off prefix */
    filename = filename + strlen(ROOT_LOC) + 1;

    n = strlen(filename);
    if (write(g_override_list_target, "/", 1) != 1 ||
            write(g_override_list_target, filename, n) != n ||
            write(g_override_list_target, "\n", 1) != 1) {
        fprintf(stderr, "Problem writing override file name: %s\n",
                strerror(errno));
        exit(1);
    }

    /* Get each subdir component, e.g.: /usr/include/linux.h does:
       move_to_ram("/usr")
       move_to_ram("/usr/include")
       move_to_ram("/usr/include/linux.h") */
    do {
        p = strpbrk(filename, "/"); if (p) *p = '\0';
        move_to_ram(basename, p ? 1 : 0);   /* Create anything but the last entry as a directory if it doesn't exist */
        if (p) {
            *p = '/';
            filename = p+1;
        }
    } while (p);
    
    return 0;
}

extern char *g_override_path;

static void decompress_and_extract_image( int fdimg ) {
    int pfds[2];
    int rc;
    pid_t fval;
    
    if (pipe(pfds) != 0) {
        fprintf(stderr, "Unable to create pipe: %s\n", strerror(errno));
        exit(1);
    }
    
    fval = fork();
    if (fval == (pid_t)-1) {
        fprintf(stderr, "Unable to fork: %s\n", strerror(errno));
    }

    /* We are the child */
    if (fval == 0) {
        /* Close unused files */
        close(pfds[1]);
        close(fdimg); 
        
        if (archive_extract(ROOT_LOC, pfds[0], 1, extract_filter) != 0) {
            fprintf(stderr, "Error extracting override archive: %s\n", strerror(errno));
            exit(1);
        }
    
        close(g_override_list_target);
        close(pfds[0]); 
        exit(0);
    }

    /* Close unused files */
    close(pfds[0]);
    close(g_override_list_target);

    /* Uncompress the file into the pipe */
    if ((rc=uncompressStream(fdimg, pfds[1])) != 0) { 
        fprintf(stderr, "BZIP2 Error: %s\n", bunzip_errors[-rc]);
        exit(1);
    }

    /* Close remaining */
    close(pfds[1]);

    /* Wait for child */
    if (waitpid(fval, &rc, 0) == -1) {
        fprintf(stderr, "Error waiting for extraction child: %s\n", strerror(errno));
        exit(1);
    }
    if (WEXITSTATUS(rc) != 0) {
        fprintf(stderr, "Extraction child exited with code %d\n", WEXITSTATUS(rc));
        exit(1);
    }
}

extern int g_root_major;
extern int g_root_minor;
void extract_override( int major, int minor, char *img ) {
    int fdin;
    char *filename;
    dev_t dt;
    struct stat st;

    printf("\nSetting up overrides from %d:%d%s\n", major,minor,img);
    
    /* If this resides on the same device as the root filesystem
       then we don't want to mount it again, just reuse the root mount */
    if (major == g_root_major && minor == g_root_minor) {
        /* NOTE: In the case of the override being on the root device,
           we take advantage of the fact that /mnt is a symlink to
           /newroot/mnt. Thus /newroot/mnt/override pointing to 
           /mnt/rootdevmnt still works */
        if (symlink(ROOT_DEV_MNT, ROOT_LOC OVR_MNT) != 0) {
            fprintf(stderr, "Can\'t link override mount to root mount: %s\n", strerror(errno));
            exit(1);
        }
    } else {
        printf("  Acquiring image...\n");
        /* Create device node and mount it */
        dt = ((major & 0xFF) << 8) | (minor & 0xFF);    /* FIXME */
        if (mknod(ROOT_LOC OVR_DEV, S_IFBLK, dt) != 0 || chmod(ROOT_LOC OVR_DEV, 0600) != 0) {
            fprintf(stderr, "Unable to create override device node: %s\n", strerror(errno));
            exit(1);
        }
        root_mkdir(OVR_MNT);
        /* NOTE: Mount is done with a path that's on the initrd but is
           symlinked to the right place.  This is done so once we
           pivot_root, /proc/mtab is still correct */
        if (!auto_fs_mount(OVR_DEV, OVR_MNT, MS_NOATIME|MS_NODIRATIME)) {
            if (!auto_fs_mount(OVR_DEV, OVR_MNT, MS_NOATIME|MS_NODIRATIME|MS_RDONLY)) {
                fprintf(stderr, "Unable to mount override device: %s\n", strerror(errno));
                exit(1);
            }
        }
    }   

    printf("  Extracting image:  ");
    fflush(stdout);

    filename = (char *)malloc(strlen(ROOT_LOC) + strlen(OVR_MNT) + strlen(img) + 1);
    sprintf(filename, ROOT_LOC OVR_MNT "%s", img);
    if ((fdin=open(filename, O_RDONLY)) == -1) {
        fprintf(stderr, "Can\'t open override image at %s: %s\n", filename, strerror(errno));
        exit(1);
    }
    free(filename);

    if (fstat(fdin, &st) != 0) {
        fprintf(stderr, "Can't stat open override file: %s\n",
                strerror(errno));
        exit(1);
    }
    if ((g_override_list_target=open(ROOT_LOC OVERRIDE_LIST, 
                    O_CREAT | O_WRONLY, 0777)) == -1 ||
            fchmod(g_override_list_target, 
                S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH) == -1) {
        fprintf(stderr, "Can't open override list: %s\n",
                strerror(errno));
        exit(1);
    }
    if (write(g_override_list_target, "0\n", 2) != 2) {
        fprintf(stderr, "Can't write time to override list: %s\n",
                strerror(errno));
        exit(1);
    }
    if (write(g_override_list_target, g_override_path, 
                strlen(g_override_path)) != strlen(g_override_path) ||
            write(g_override_list_target, "\n", 1) != 1) {
        fprintf(stderr, "Can't write override path: %s\n",
                strerror(errno));
        exit(1);
    }
    if (st.st_size == 0) {
        close(g_override_list_target);
        printf("Zero size.  Done");
    } else {
        decompress_and_extract_image(fdin); 
    }

    close(fdin);
    printf("\n");

    printf("Done setting up override\n\n");
}
